#include <iostream>
#include <fstream>
#include <vector>
#include <string>
#include <sstream>
#include <cmath>

using namespace std;

/* --------------------------------------------------------------------
 *  Platform/build specific symbols
 * --------------------------------------------------------------------
 */
#if defined(_WIN32)
  #if ((_MSC_VER == 1400) || (_MSC_VER == 1500))
    //---- WINDOWS MSVC COMPILER INSISTS THAT SECURE STRING FNS BE USED.
    #define HAVE_MSVC_SECURE_STRING_FNS
  #endif
#endif

/* --------------------------------------------------------------------
 *  Include file (WELLPOS.INC) editor
 * --------------------------------------------------------------------
 */
static bool writeNewIncludeFile(vector<double> wellPos)
{
    ofstream outfile ("WELLPOS.INC");

    // WRITING WELSPECS DATA
    outfile << "WELSPECS" << endl;
    outfile << "INJ1\tG1\t" << wellPos[0] << "\t" << wellPos[1]
        << "\t1715\tWAT /" << endl;
    outfile << "PROD1\tG2\t" << wellPos[2] << "\t" << wellPos[3]
        << "\t1715\tOIL /" << endl;
    outfile << "PROD2\tG2\t" << wellPos[4] << "\t" << wellPos[5]
        << "\t1715\tOIL /" << endl;
    outfile << "PROD3\tG2\t" << wellPos[6] << "\t" << wellPos[7]
        << "\t1715\tOIL /" << endl;
    outfile << "PROD4\tG2\t" << wellPos[8] << "\t" << wellPos[9]
        << "\t1715\tOIL /" << endl;
    outfile << "/" << endl;

    // WRITING WELSPECS DATA
    outfile << "COMPDAT" << endl;
    outfile << "INJ1\t" << wellPos[0] << "\t" << wellPos[1]
        << "\t1\t1\tOPEN\t1\t1*\t0.1905\t4*\t1* /" << endl;
    outfile << "PROD1\t" << wellPos[2] << "\t" << wellPos[3]
        << "\t1\t1\tOPEN\t1\t1*\t0.1905\t4*\t1* /" << endl;
    outfile << "PROD2\t" << wellPos[4] << "\t" << wellPos[5]
        << "\t1\t1\tOPEN\t1\t1*\t0.1905\t4*\t1* /" << endl;
    outfile << "PROD3\t" << wellPos[6] << "\t" << wellPos[7]
        << "\t1\t1\tOPEN\t1\t1*\t0.1905\t4*\t1* /" << endl;
    outfile << "PROD4\t" << wellPos[8] << "\t" << wellPos[9]
        << "\t1\t1\tOPEN\t1\t1*\t0.1905\t4*\t1* /" << endl;
    outfile << "/" << endl;

    outfile.close();

    return( 0 );
}

/* --------------------------------------------------------------------
 *  ECLIPSE simulation launcher
 * --------------------------------------------------------------------
 */

static bool runEclipseSim()
{
    remove("SIMPLECASE.RSM");

    system("$eclipse SIMPLECASE");

    return( 0 );
}



/* --------------------------------------------------------------------
 *  RSM file reader that extracts last FOPT value
 * --------------------------------------------------------------------
 */

double readRSMFile(const string rsmFileName)
{

    double foptLast = -1;

    ifstream file(rsmFileName + ".RSM");

    if (file)
    {
        string lastLine;
        while(file >> ws && getline(file, lastLine))
        {

        }
        string buf;
        stringstream ss(lastLine);
        vector<string> tokens;
        while (ss >> buf)
        {
            tokens.push_back(buf);
        }
        if (!tokens.empty())
        {
            foptLast = stod(tokens.end()[-2]); // Get second to last element
        }
    }
    else
    {
        cout << "Unable to open .RSM file." << endl;
    }

    return( foptLast );

}

/* --------------------------------------------------------------------
 *  Keep track of a single run and record it to a file
 * --------------------------------------------------------------------
 */

static bool writeSummaryToFile(vector<double> wellPos,
                               double foptLast)
{

    ofstream sumfile ("RUNSUM.TXT", std::ios::app);

    sumfile << "============ NEW ITERATION ============" << endl;
    sumfile << "=== WELL COORDINATES ===" << endl;
    sumfile << "Well\tX\tY" << endl;
    sumfile << "INJ1\t" << wellPos[0] << "\t" << wellPos[1] << endl;
    sumfile << "PROD1\t" << wellPos[2] << "\t" << wellPos[3] << endl;
    sumfile << "PROD2\t" << wellPos[4] << "\t" << wellPos[5] << endl;
    sumfile << "PROD3\t" << wellPos[6] << "\t" << wellPos[7] << endl;
    sumfile << "PROD4\t" << wellPos[8] << "\t" << wellPos[9] << endl;
    sumfile << "=== FOPT ===" << endl;
    sumfile << "FOPT = " << foptLast << endl;
    sumfile << "============ END ITERATION ============" << endl;
    sumfile << "" << endl;

    sumfile.close();

    return( 0 );

}

/* --------------------------------------------------------------------
 *  Internal Function read_input_file
 * --------------------------------------------------------------------
 */

static bool  readInputFile (const string &szExeName,
                            const string &szFileName,
                            vector<double> &wellPos)
{
    const int varNum = 10; // Declare the number of variables
    const int REQBUFLEN = 10;
    char reqBuf[REQBUFLEN + 1];

    ifstream fp;

    //---- OPEN THE INPUT FILE.
    fp.open (szFileName.c_str());
    if (!fp)
    {
        cerr << szExeName << " - ERROR opening input file '"
             << szFileName << "'." << endl;
        return( -1 );
    }

    //---- READ AND VERIFY THE INPUT REQUEST TYPE.
    fp >> reqBuf;
    for (int  i = 0; i < REQBUFLEN + 1; i++)
    {
        if ((reqBuf[i] == '\n') || (reqBuf[i] == '\r'))
        {
            reqBuf[i] = 0;
            break;
        }
    }
    if ((reqBuf[0] != 'F') || (reqBuf[1] != 0))
    {
        cerr << szExeName << " - ERROR reading request type from '"
             << szFileName << "'." << endl;
        cerr << "  Read '" << reqBuf << "'" << endl;
        fp.close();
        return( -1 );
    }

    //---- READ THE LENGTH OF x.
    int  n;
    fp >> n;
    if (n != varNum)
    {
        cerr << szExeName << " - ERROR reading n from '"
             << szFileName << "'." << endl;
        cerr << "  Read " << n << ", but problem size should be " << varNum << endl;
        fp.close();
        return( -1 );
    }

    wellPos.resize (n);

    //---- READ x.
    for (int  i = 0; i < n; i++)
    {
        fp >> wellPos[i];
        cout << round(wellPos[i]) << "\t";
    }
    cout << "" << endl;
    fp.close();

    //---- RETURN SUCCESS.
    return( 0 );
}

/* --------------------------------------------------------------------
 *  Internal Function write_output_file
 * --------------------------------------------------------------------
 */
static bool  writeOutputFile (const string & szExeName,
                              const string & szFileName,
                              const double f)
{
    ofstream  fp;

    fp.open (szFileName.c_str(), ios::trunc);
    if (!fp)
    {
        cerr << szExeName << " - ERROR opening output file '"
             << szFileName << "'." << endl;
        return( -1 );
    }

    //---- WRITE THE NUMBER OF OBJECTIVES AND THEIR VALUES TO THE OUTPUT.
    fp << "1" << endl;
    fp.precision (15);     //-- WRITE ALL DIGITS
    fp << f << endl;

    fp.close();

    return( 0 );
}

/* --------------------------------------------------------------------
 *  Main routine for evaluation executable.
 *
 *  Each execution reads a vector from a file, evaluates the objective
 *  function, and writes the function value to an output file.  HOPSPACK
 *  generates the file names and decides when to invoke this executable.
 *  If there is an error, either create no output file, or create an
 *  output file with an error message.
 *
 *  @param argc  Number of command line arguments.
 *  @param argv  Command line arguments (input and output file names).
 *  @return      0 if successful.
 * --------------------------------------------------------------------
 */
int  main (const int argc,
           const char * const argv[])
{
    vector<double> wellPos;
    double foptLast = -1;
    int nRetStatus;

    /* CHECK THE COMMAND LINE ARGUMENTS.
     * IN THIS EXAMPLE ONLY THE INPUT AND OUTPUT FILE NAMES ARE USED.
     */
    if (argc != 5)
    {
        fprintf (stderr, "usage: %s <input file> <output file> <tag> <type>\n",
                 argv[0]);
        return( -100 );
    }

    // Read an internal input file
    nRetStatus = readInputFile (argv[0], argv[1], wellPos);
    if (nRetStatus != 0)
        return( nRetStatus );

    for (unsigned int  i = 0; i < wellPos.size(); i++)
    {
        wellPos[i] = round(wellPos[i]);
    }

    // Using wellPos values to write a new WELLPOS.INC file
    nRetStatus = writeNewIncludeFile(wellPos);
    if (nRetStatus != 0)
        return( nRetStatus );

    do
    {
        // Run ECLIPSE simulations
        nRetStatus = runEclipseSim();
        if (nRetStatus != 0)
            return( nRetStatus );

        // Reading RSM file to extract last FOPT value
        foptLast = readRSMFile("SIMPLECASE");
        if (foptLast != -1)
        {
            // Write summary of a run to a file
            nRetStatus = writeSummaryToFile(wellPos, foptLast);
            if (nRetStatus != 0)
                return( nRetStatus );
        }
    } while( foptLast == -1 );

    // Write an internal output file
    nRetStatus = writeOutputFile (argv[0], argv[2], foptLast);

    return( nRetStatus );
}
